(* type var *)
type var = (Strng.t * int)
(* type t *)
type t = Var of var | Abs of (var * t) | App of (t * t)
(* var : char -> t *)
let var s =  Var(Strng.of_string s,0)
(* abs : char list -> t -> t *)
let abs xs body = Lst.foldr (fun x t ->
 Abs((Strng.of_string x,0),t)) xs body
(* app : t list -> t *)
let app ts = Lst.foldl1 (fun s t -> App(s,t)) ts
(* max_index : t -> int *)
let rec max_index = function
 | Var(_,i)     -> i
 | Abs((_,i),u) -> max i (max_index u)
 | App(u,v)     -> max (max_index u) (max_index v)
(* rename_bound : int -> t -> t *)
let rename_bound i t =
 let inc (id,i) j = (id,i+j) in
 let rec rename_bound i bvs = function
  | Var x as v ->
   if St.mem x bvs then Var(inc x i) else v
  | Abs(x,u)  ->
   Abs(inc x i,rename_bound i (St.insert x bvs) u)
  | App(u,v)  ->
   App(rename_bound i bvs u,rename_bound i bvs v)
 in
 rename_bound i St.empty t
(* substitute : var -> t -> t *)
let rec substitute x t = function
 | Var y as v    -> if y = x then t else v
 | Abs(y,u) as s -> if y = x then s
                             else Abs(y,substitute x t u)
 | App(u,v)      -> App(substitute x t u,substitute x t v)
(* beta : t -> t *)
let beta = function
 | App(Abs(x,u),v) ->
  substitute x v (rename_bound (max_index v + 1) u)
 | _               -> failwith "not a redex"
(* is_redex : t -> bool *)
let is_redex = function
 | App(Abs _,_) -> true
 | _ -> false
(* outer : (t -> bool) -> t -> t *)
let outer stop t =
 let rec outer t = if stop t then (false,t) else match t with
  | Var _ as x -> (false,x)
  | Abs(x,u) -> let (r,u) = outer u in (r,Abs(x,u))
  | App(u,v) as t ->
   if is_redex t then
    (true,beta t)
   else
    let (r,u) = outer u in
    if r then (r,App(u,v))
         else let (r,v) = outer v in (r,App(u,v))
 in snd (outer t)
(* inner : (t -> bool) -> t -> t *)
let inner stop t =
 let rec inner t = if stop t then (false,t) else match t with
  | Var _ as x -> (false,x)
  | Abs(x,u) -> let (r,u) = inner u in (r,Abs(x,u))
  | App(u,v) ->
   let (r,u) = inner u in
   if r then (r,App(u,v)) else let (r,v) = inner v in
   if r then (r,App(u,v)) else let t = App(u,v) in
   if is_redex t then (true,beta t) else (false,t)
 in snd (inner t)
(* vars : t -> var St.t *)
let rec vars = function
 | Var x    -> St.singleton x
 | Abs(x,u) -> St.insert x (vars u)
 | App(u,v) -> St.union (vars u) (vars v)
(* fvars : t -> var St.t *)
let rec fvars = function
 | Var x    -> St.singleton x
 | Abs(x,u) -> St.diff (fvars u) (St.singleton x)
 | App(u,v) -> St.union (fvars u) (fvars v)
(* bvars : t -> var St.t *)
let rec bvars = function
 | Var _    -> St.empty
 | Abs(x,u) -> St.insert x (bvars u)
 | App(u,v) -> St.union (bvars u) (bvars v)
(* is_app : t -> bool *)
let is_app = function App _ -> true
                    | _     -> false
(* is_whnf : t -> bool *)
let rec is_whnf = function App _ -> false
                         | _     -> true
(* reducts : t -> t list *)
let rec reducts = function
 | Var _         -> []
 | Abs(x,u)      -> Lst.map (fun t -> Abs(x,t)) (reducts u)
 | App(u,v) as t ->
  let us = Lst.map (fun l -> App(l,v)) (reducts u) in
  let vs = Lst.map (fun r -> App(u,r)) (reducts v) in
  if is_redex t then beta t :: us @ vs
                else us @ vs
(* is_nf : t -> bool *)
let rec is_nf = function
 | Var _         -> true
 | Abs(x,u)      -> is_nf u
 | App(u,v) as t -> if is_redex t then false
                                  else is_nf u && is_nf v
(* is_closed : t -> bool *)
let is_closed t = St.is_empty (fvars t)
(* var2strng : var -> Strng.t *)
let var2strng (id,i) = if i < 0 then id else id @ (Strng.of_int i)
(* to_string : t -> Strng.t *)
let rec to_strng = function
 | Var x -> var2strng x
 | Abs(x,u) -> '('::'%'::var2strng x @ ('.'::to_strng u) @ [')']
 | App(u,v) -> '('::to_strng u @ (' '::to_strng v) @ [')']
(* sub : t -> t St.t *)
let rec sub = function
 | Var _ as x     -> St.singleton x
 | Abs(x,u) as t -> St.insert t (sub u)
 | App(u,v) as t -> St.insert t (St.union (sub u) (sub v))
(* sub_list : t -> t list *)
let rec sub_list = function
 | Var _ as x    -> [x]
 | Abs(x,u) as t -> t::sub_list u
 | App(u,v) as t -> t::sub_list u @ sub_list v
(* redexes : t -> t St.t *)
let redexes t = St.filter (is_redex) (sub t)
(* redexes_list : t -> t list *)
let redexes_list t = Lst.filter (is_redex) (sub_list t)
(* by_name : t -> t *)
let by_name t =
 let u = Lst.until is_whnf (outer is_whnf) t in
 Lst.until is_nf (outer is_nf) u
(* by_value : t -> t *)
let by_value t =
 let u = Lst.until is_whnf (inner is_whnf) t in
 Lst.until is_nf (inner is_nf) u
(* pvar : Format.formatter -> var -> unit *)
let pvar fmt (id,i) =
 let rec pid fmt = function
  | []    -> ()
  | c::cs -> Format.fprintf fmt "%c%a" c pid cs
 in
 if i = 0 then Format.fprintf fmt "%a" pid id
          else Format.fprintf fmt "%a%i" pid id i
(* print : Format.formatter -> t -> unit *)
let print fmt t =
 let rec pabs fmt = function
  | Abs(x,(Abs _ as u)) -> Format.fprintf fmt "%a%a" pvar x pabs u
  | Abs(x,u) -> Format.fprintf fmt "%a.%a" pvar x print u
  | t -> print fmt t
 and print fmt = function
  | Var x -> pvar fmt x
  | Abs _ as t -> Format.fprintf fmt "\\%a" pabs t
  | App(Var _ as u,(Var _ as v))
  | App(App _ as u,(Var _ as v)) ->
   Format.fprintf fmt "%a %a" print u print v
  | App(Var _ as u,(Abs _ as v))
  | App(Var _ as u,(App _ as v))
  | App(App _ as u,(Abs _ as v))
  | App(App _ as u,(App _ as v)) ->
   Format.fprintf fmt "%a (%a)" print u print v
  | App (Abs _ as u,(Var _ as v)) ->
   Format.fprintf fmt "(%a) %a" print u print v
  | App(Abs _ as u,(Abs _ as v))
  | App(Abs _ as u,(App _ as v)) ->
   Format.fprintf fmt "(%a) (%a)" print u print v
 in
 print fmt t
(* toplevel_printer : Format.formatter -> t -> unit *)
let toplevel_printer fmt = Format.fprintf fmt "%a%!" print
(* toplevel_printer_term_set : Format.formatter -> t St.t -> unit *)
let toplevel_printer_term_set = St.fprintf toplevel_printer
(* toplevel_printer_var_set : Format.formatter -> var St.t -> unit *)
let toplevel_printer_var_set = St.fprintf pvar
(*  *)
